/*
 * (c) 2023-2024 Jens Mueller
 *
 * JKCLOAD
 *
 * Fenster zur Anzeige der Hilfe
 */

package jkcload.ui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.net.URL;
import java.util.Properties;
import javax.swing.JButton;
import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JToolBar;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;
import jkcload.Main;


public class HelpFrm extends JFrame implements
					ActionListener,
					HyperlinkListener,
					MenuListener
{
  private static final String HELP_PAGE_RESOURCE = "/help/jkcload.htm";
  private static final String PROP_PREFIX        = Main.PROP_PREFIX + "help.";

  private boolean     notified;
  private JButton     btnBack;
  private JMenu       mnuEdit;
  private JMenuItem   mnuClose;
  private JMenuItem   mnuCopy;
  private JMenuItem   mnuSelectAll;
  private JEditorPane editorPane;


  public HelpFrm()
  {
    this.notified = false;
    setDefaultCloseOperation( JFrame.DISPOSE_ON_CLOSE );
    setTitle( Main.APPNAME + " Hilfe" );


    // Fenster-Icons
    UIUtil.setIconImagesAt( this );


    // Menu
    JMenu mnuFile = new JMenu( "Datei" );
    mnuFile.setMnemonic( KeyEvent.VK_D );

    this.mnuClose = new JMenuItem( "Schlie\u00DFen" );
    mnuFile.add( this.mnuClose );

    this.mnuEdit = new JMenu( "Bearbeiten" );
    mnuEdit.setMnemonic( KeyEvent.VK_B );

    this.mnuCopy = new JMenuItem( "Kopieren" );
    this.mnuEdit.add( this.mnuCopy );
    this.mnuEdit.addSeparator();

    this.mnuSelectAll = new JMenuItem( "Alles ausw\u00E4hlen" );
    mnuEdit.add( this.mnuSelectAll );

    JMenuBar mnuBar = new JMenuBar();
    mnuBar.add( mnuFile );
    mnuBar.add( this.mnuEdit );
    setJMenuBar( mnuBar );


    // Fensterinhalt
    setLayout( new BorderLayout() );

    JToolBar toolBar = new JToolBar();
    toolBar.setFloatable( false );
    toolBar.setRollover( true );
    add( toolBar, BorderLayout.NORTH );

    this.btnBack = UIUtil.createImageButton(
					"/images/back.png",
					"Zur\u00FCck" );
    toolBar.add( this.btnBack );

    this.editorPane = new JEditorPane();
    this.editorPane.setEditable( false );
    this.editorPane.setPreferredSize( new Dimension( 500, 400 ) );
    setStartPage();
    add( new JScrollPane( editorPane ), BorderLayout.CENTER );


    // Sonstiges
    if( !UIUtil.restoreWindowBounds(
			this,
			Main.getProperties(),
			PROP_PREFIX ) )
    {
      pack();
      setLocationByPlatform( true );
    }
  }


  public void doClose()
  {
    setVisible( false );
    dispose();
  }


  public void getSettings( Properties props )
  {
    UIUtil.getWindowBounds( this, props, PROP_PREFIX );
  }


	/* --- ActionListener --- */

  @Override
  public void actionPerformed( ActionEvent e )
  {
    Object src = e.getSource();
    if( src == this.btnBack ) {
      setStartPage();
    } else if( src == this.mnuClose ) {
      doClose();
    } else if( src == this.mnuCopy ) {
      this.editorPane.copy();
    } else if( src == this.mnuSelectAll ) {
      this.editorPane.selectAll();
      this.editorPane.requestFocus();
    }
  }


	/* --- HyperlinkListener --- */

  @Override
  public void hyperlinkUpdate( HyperlinkEvent e )
  {
    if( e.getEventType() == HyperlinkEvent.EventType.ACTIVATED ) {
      try {
	this.editorPane.setPage( e.getURL() );
	this.btnBack.setEnabled( true );
      }
      catch( IOException ex ) {
	setStartPage();
      }
    }
  }


	/* --- HyperlinkListener --- */

  @Override
  public void menuCanceled( MenuEvent e )
  {
    // leer
  }


  @Override
  public void menuDeselected( MenuEvent e )
  {
    // leer
  }


  @Override
  public void menuSelected( MenuEvent e )
  {
    if( e.getSource() == this.mnuEdit ) {
      int begPos = this.editorPane.getSelectionStart();
      this.mnuCopy.setEnabled(
	(begPos >= 0) && (begPos < this.editorPane.getSelectionEnd()) );
    }
  }


	/* --- ueberschriebene Methoden --- */

  @Override
  public void addNotify()
  {
    super.addNotify();
    if( !this.notified ) {
      this.notified = true;
      this.btnBack.addActionListener( this );
      this.mnuClose.addActionListener( this );
      this.mnuCopy.addActionListener( this );
      this.mnuSelectAll.addActionListener( this );
      this.mnuEdit.addMenuListener( this );
      this.editorPane.addHyperlinkListener( this );
    }
  }


  @Override
  public void removeNotify()
  {
    super.removeNotify();
    if( this.notified ) {
      this.notified = false;
      this.btnBack.removeActionListener( this );
      this.mnuClose.removeActionListener( this );
      this.mnuCopy.removeActionListener( this );
      this.mnuSelectAll.removeActionListener( this );
      this.mnuEdit.removeMenuListener( this );
      this.editorPane.removeHyperlinkListener( this );
    }
  }


	/* --- private Methoden --- */

  private void setStartPage()
  {
    try {
      URL url = getClass().getResource( HELP_PAGE_RESOURCE );
      if( url != null ) {
	editorPane.setPage( url );
	this.btnBack.setEnabled( false );
      }
    }
    catch( IOException ex ) {}
  }
}
